/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.core.resolvers;

import net.craftforge.essential.supply.Consumer;
import net.craftforge.essential.supply.exceptions.BadInputException;

import java.io.InputStream;
import java.io.UnsupportedEncodingException;

/**
 * A body resolver provides a convenient interface to deserialize the content
 * of an HTTP request or response body input stream.
 *
 * @author Christian Bick
 * @since 29.11.11
 */
public interface BodyResolver {

    /**
     * Gets the body input stream.
     *
     * @return The body input stream
     */
    InputStream getBodyInputStream();

    /**
     * Gets the body content as its String representation using the given charset
     * for decoding the byte stream.
     *
     * @param charset The charset
     * @return The String representation of the body content
     * @throws BadInputException if the body content is not valid
     * @throws UnsupportedEncodingException if the given charset is not supported
     */
    String getBody(String charset) throws BadInputException, UnsupportedEncodingException;

    /**
     * Gets the body content as its object of the given class representation using the given consumer for
     * deserialization and assuming the byte stream to be encoded in the given charset.
     *
     * @param consumer The consumer
     * @param clazz The object's class
     * @param charset The charset
     * @return The body object
     * @throws BadInputException if the body content is not valid
     * @throws UnsupportedEncodingException if the given charset is not supported
     */
    Object getBodyAsObject(Consumer consumer, Class<?> clazz, String charset) throws BadInputException, UnsupportedEncodingException;
}
