/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.core.resolvers;

import net.craftforge.essential.core.exceptions.MissingHeaderException;
import net.craftforge.essential.core.utils.HeaderUtils;

import javax.inject.Inject;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

/**
 * {@inheritDoc}
 *
 * @author Christian Bick
 * @since 21.08.11
 */
public class DefaultHeaderResolver implements HeaderResolver {

    private HeaderMap headerMap;

    @Inject
    public DefaultHeaderResolver(HeaderMap headerMap) {
        this.headerMap = headerMap;
    }

    /**
     * {@inheritDoc}
     */
    public String getHeaderValue(String headerName) throws MissingHeaderException {
        return getHeaderValue(headerName, null);
    }

    /**
     * {@inheritDoc}
     */
    public String getHeaderValue(String headerName, String defaultValue) throws MissingHeaderException {
        String[] headerValues = headerMap.get(headerName);
        if (headerValues != null && headerValues.length > 0) {
            return HeaderUtils.concatenateHeaderValues(headerValues);
        } else if (defaultValue != null) {
            return defaultValue;
        }
        throw getMissingHeaderException(headerName);
    }

    /**
     * {@inheritDoc}
     */
    public String[] getHeaderValues(String headerName) throws MissingHeaderException {
        return getHeaderValues(headerName, null);
    }

    /**
     * {@inheritDoc}
     */
    public String[] getHeaderValues(String headerName, String[] defaultValues) throws MissingHeaderException {
        String[] headerValues = headerMap.get(headerName);
        if (headerValues != null) {
            List<String> splitHeaderValues = new LinkedList<String>();
            for (String headerValue : headerValues) {
                String[] split = HeaderUtils.splitHeaderValues(headerValue);
                splitHeaderValues.addAll(Arrays.asList(split));
            }
            String[] array = new String[splitHeaderValues.size()];
            return splitHeaderValues.toArray(array);
        } else if (defaultValues != null) {
            return defaultValues;
        }
        throw getMissingHeaderException(headerName);
    }

    /**
     * {@inheritDoc}
     */
    public String getHeaderValueAttribute(String headerName, String attributeName) throws MissingHeaderException {
        String headerValue = getHeaderValue(headerName);
        String attributeValue = HeaderUtils.extractHeaderValueAttribute(headerValue, attributeName);
        if (attributeValue != null) {
            return attributeValue;
        }
        throw new MissingHeaderException();
    }

    /**
     * {@inheritDoc}
     */
    public String getHeaderValueAttribute(String headerName, String attributeName, String defaultValue) throws MissingHeaderException {
        String defaultHeaderValue = ";" + attributeName + "=" + defaultValue;
        String headerValue = getHeaderValue(headerName, defaultHeaderValue);
        String attributeValue = HeaderUtils.extractHeaderValueAttribute(headerValue, attributeName);
        if (attributeValue != null) {
            return attributeValue;
        } else if (defaultValue != null) {
            return defaultValue;
        }
        throw new MissingHeaderException();
    }

    protected MissingHeaderException getMissingHeaderException(String headerName) {
        return new MissingHeaderException("The header " + headerName + " is " +
                    "expected to be part of the request but is missing");
    }

}
