/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.core.resolvers;

import net.craftforge.essential.core.exceptions.MissingParameterException;
import net.craftforge.essential.supply.exceptions.BadInputException;
import net.craftforge.essential.supply.suppliers.basic.BasicConsumer;
import net.craftforge.reflection.utils.ReflUtils;

import javax.inject.Inject;
import java.io.StringReader;
import java.util.Collection;

/**
 * {@inheritDoc}
 *
 * @author Christian Bick
 * @since 21.08.11
 */
public class DefaultParamResolver implements ParamResolver {

    private ParamMap paramMap;

    @Inject
    public DefaultParamResolver(ParamMap paramMap) {
        this.paramMap = paramMap;
    }

    /**
     * {@inheritDoc}
     */
    public String getParameterValue(String parameterName) throws MissingParameterException {
        return getParameterValue(parameterName, null);
    }

    /**
     * {@inheritDoc}
     */
    public String getParameterValue(String parameterName, String defaultValue) throws MissingParameterException {
        String[] parameterValues = paramMap.get(parameterName);
        if (parameterValues != null && parameterValues.length > 0) {
            return parameterValues[0];
        } else if (defaultValue != null) {
            return defaultValue;
        }
        throw getMissingParameterException(parameterName);
    }

    /**
     * {@inheritDoc}
     */
    public String[] getParameterValues(String parameterName) throws MissingParameterException {
        return getParameterValues(parameterName, null);
    }

    /**
     * {@inheritDoc}
     */
    public String[] getParameterValues(String parameterName, String[] defaultValues) throws MissingParameterException {
        String[] parameterValues = paramMap.get(parameterName);
        if (parameterValues != null) {
            return parameterValues;
        } else if (defaultValues != null) {
            return defaultValues;
        }
        throw getMissingParameterException(parameterName);
    }

    /**
     * {@inheritDoc}
     */
    public Object getParameterValueAsObject(Class<?> clazz, String parameterName) throws BadInputException {
        return getParameterValueAsObject(clazz, parameterName, null);
    }

    /**
     * {@inheritDoc}
     */
    public Object getParameterValueAsObject(Class<?> clazz, String parameterName, String defaultValue) throws BadInputException {
        String parameterValue = getParameterValue(parameterName, defaultValue);
        return convertToObject(clazz, parameterValue);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public <G> Collection<G> getParameterValuesAsObjects(Class<?> collectionType, Class<G> typeArgument, String parameterName) throws BadInputException {
        return getParameterValuesAsObjects(collectionType, typeArgument, parameterName, null);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public <G> Collection<G> getParameterValuesAsObjects(Class<?> collectionType, Class<G> typeArgument, String parameterName, String[] defaultValues) throws BadInputException {
        Collection<G> collection = ReflUtils.createCollection(collectionType);
        for (String paramValue : getParameterValues(parameterName, defaultValues)) {
            collection.add((G) convertToObject(typeArgument, paramValue));
        }
        return collection;
    }

    /**
     * Converts a parameter value to and object of the specified
     * class.
     *
     * @param clazz The object's class
     * @param parameterValue The parameter value
     * @return The object
     * @throws BadInputException if the parameter is not present while no default is given
     * or the parameter value cannot be deserialized
     */
    protected Object convertToObject(Class<?> clazz, String parameterValue) throws BadInputException {
        return new BasicConsumer().consume(clazz, new StringReader(parameterValue));
    }

    protected MissingParameterException getMissingParameterException(String paramName) {
        return new MissingParameterException("The query parameter " + paramName + " is " +
                    "expected to be passed in the request but is missing");
    }

}
