/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.core.resolvers;

import net.craftforge.essential.core.exceptions.MissingHeaderException;

/**
 * A header resolver provides a convenient interface to access the values of
 * HTTP request or response headers.
 *
 * @author Christian Bick
 * @since 29.11.11
 */
public interface HeaderResolver {

    /**
     * Gets a single header value of a a header with the given name.
     *
     * @param headerName The header name
     * @return The header values
     * @throws net.craftforge.essential.core.exceptions.MissingHeaderException if the header is not present
     */
    String getHeaderValue(String headerName) throws MissingHeaderException;

    /**
     * Gets a single header value of a a header with the given name using the given
     * default value if the header is not present.
     *
     * @param headerName The header name
     * @param defaultValue The default value
     * @return The header value
     * @throws net.craftforge.essential.core.exceptions.MissingHeaderException if the header is not present and the default value is null
     */
    String getHeaderValue(String headerName, String defaultValue) throws MissingHeaderException;

    /**
     * Gets all header values of headers with the given name.
     *
     * @param headerName The header name
     * @return The header values
     * @throws net.craftforge.essential.core.exceptions.MissingHeaderException if the header is not present
     */
    String[] getHeaderValues(String headerName) throws MissingHeaderException;

    /**
     * Gets all header values of headers with the given name using the given default values
     * if the header is not present.
     *
     * @param headerName The header name
     * @param defaultValues The header default values
     * @return The header values
     * @throws net.craftforge.essential.core.exceptions.MissingHeaderException if the header is not present and the default value is null
     */
    String[] getHeaderValues(String headerName, String[] defaultValues) throws MissingHeaderException;

    /**
     * Gets the header attribute with the given attribute name from the header value
     * with the given header name.
     *
     * @param headerName The header name
     * @param attributeName The attribute name
     * @return The header value attribute
     * @throws net.craftforge.essential.core.exceptions.MissingHeaderException if the header is not present or the attribute is not specified
     */
    String getHeaderValueAttribute(String headerName, String attributeName) throws MissingHeaderException;

    /**
     * Gets the header attribute with the given attribute name from the header value
     * with the given header name. Uses the given default value if the header is not
     * present or the attribute is not specified.
     *
     * @param headerName The header name
     * @param attributeName The attribute name
     * @param defaultValue The default value
     * @return The header value attribute
     * @throws net.craftforge.essential.core.exceptions.MissingHeaderException if the header is not present or the attribute is not specified
     * and the default value is null
     */
    String getHeaderValueAttribute(String headerName, String attributeName, String defaultValue) throws MissingHeaderException;
}
