/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.core.resolvers;

import net.craftforge.essential.core.exceptions.MissingParameterException;
import net.craftforge.essential.supply.exceptions.BadInputException;

import java.util.Collection;

/**
 * A parameter resolver provides a convenient interface to access the values of
 * HTTP request parameters.
 *
 * @author Christian Bick
 * @since 29.11.11
 */
public interface ParamResolver {

    /**
     * Gets the parameter value of a parameter with the specified
     * parameter name.
     *
     * @param parameterName The parameter name
     * @return The parameter value
     * @throws net.craftforge.essential.core.exceptions.MissingParameterException if the parameter is not present and no default value is given
     */
    String getParameterValue(String parameterName) throws MissingParameterException;

    /**
     * Gets the parameter value of a parameter with the specified
     * parameter name using the given default value if no parameter
     * with that name is present.
     *
     * @param parameterName The parameter name
     * @param defaultValue The default value
     * @return The parameter value
     * @throws net.craftforge.essential.core.exceptions.MissingParameterException if the parameter is not present and no default value is given
     */
    String getParameterValue(String parameterName, String defaultValue) throws MissingParameterException;

    /**
     * Gets the parameter values of a parameter with the specified
     * parameter name.
     *
     * @param parameterName The parameter name
     * @return The parameter values
     * @throws net.craftforge.essential.core.exceptions.MissingParameterException if the parameter is not present and no default values are given
     */
    String[] getParameterValues(String parameterName) throws MissingParameterException;

    /**
     * Gets the parameter values of a parameter with the specified
     * parameter name using the given default values if no parameter
     * with that name is present.
     *
     * @param parameterName The parameter name
     * @param defaultValues The default values
     * @return The parameter values
     * @throws net.craftforge.essential.core.exceptions.MissingParameterException if the parameter is not present and no default values are given
     */
    String[] getParameterValues(String parameterName, String[] defaultValues) throws MissingParameterException;

    /**
     * Gets the parameter value of a parameter with the specified
     * parameter name as an object of the given class.
     *
     * @param clazz The object's class (a primitive or virtual primitive)
     * @param parameterName The parameter name
     * @return The value as an object
     * @throws net.craftforge.essential.supply.exceptions.BadInputException if the parameter is not present or the parameter value cannot be deserialized
     */
    Object getParameterValueAsObject(Class<?> clazz, String parameterName) throws BadInputException;

    /**
     * Gets the parameter value of a parameter with the specified
     * parameter name as an object of the given class. Uses the given
     * default value if no parameter with that name is present.
     *
     * @param clazz The object's class (a primitive or virtual primitive)
     * @param parameterName The parameter name
     * @param defaultValue The default value
     * @return The value as an object
     * @throws net.craftforge.essential.supply.exceptions.BadInputException if the parameter is not present and no default value is given
     * or the parameter value cannot be deserialized
     */
    Object getParameterValueAsObject(Class<?> clazz, String parameterName, String defaultValue) throws BadInputException;

    /**
     * Gets the input for a collection type using the generic type as deserialization hint
     * for the parameter values associated to the given parameter name.
     *
     * @param collectionType The collection type
     * @param typeArgument The collection's generic type argument
     * @param parameterName The parameter name
     * @param <G> The generic type parameter
     * @return The collection initialized from the parameter values
     * @throws net.craftforge.essential.supply.exceptions.BadInputException if the parameter is not present or the parameter value cannot be deserialized
     */
    @SuppressWarnings("unchecked")
    <G> Collection<G> getParameterValuesAsObjects(Class<?> collectionType, Class<G> typeArgument, String parameterName) throws BadInputException;

    /**
     * Gets the input for a collection type using the generic type as deserialization hint
     * for the parameter values associated to the given parameter name.
     *
     * @param collectionType The collection type
     * @param typeArgument The collection's generic type argument
     * @param parameterName The parameter name
     * @param defaultValues The default values to use if no parameter values are given
     * @param <G> The generic type parameter
     * @return The collection initialized from the parameter values
     * @throws net.craftforge.essential.supply.exceptions.BadInputException if the parameter is not present and no default values are given
     * or the parameter value cannot be deserialized
     */
    @SuppressWarnings("unchecked")
    <G> Collection<G> getParameterValuesAsObjects(Class<?> collectionType, Class<G> typeArgument, String parameterName, String[] defaultValues) throws BadInputException;
}
