/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.core.utils;

import net.craftforge.commons.collections.CaseInsensitiveMap;
import org.apache.commons.codec.binary.Base64;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Utilities for header actions.
 *
 * @author Christian Bick
 * @since 29.08.11
 */
public class HeaderUtils {

    /**
     * Finds all charset=XYZ notations regardless of whitespace characters and
     * provides a group for XYZ
     */
    public static final Pattern FIND_CHARSET = Pattern.compile("charset\\s*=\\s*(\\S+)");

    /**
     * Finds a character sequence that does not contain an equal character.
     */
    public static final Pattern FIND_NO_EQUAL_CONTAINED = Pattern.compile("^[^=]+$");

    /**
     * Finds everything from a semi colon to a comma (including both)
     */
    public static final Pattern FIND_SEMI_COLON_TO_COMMA = Pattern.compile(";.*,");

    /**
     * Finds everything from a semi colon to end of string (including semi colon)
     */
    public static final Pattern FIND_SEMI_COLON_TO_END = Pattern.compile("[ ]*;.*$");

    /**
     * Finds header attributes and provides groups for the attribute name and value
     */
    public static final Pattern FIND_ATTRIBUTES = Pattern.compile(";([a-zA-Z0-9]+)=([^ ,;]+)");

    /**
     * Gets an HTTP basic authorization header value from a username and password.
     *
     * @param username The username
     * @param password The password
     * @return The HTTP basic authorization header
     */
    public static String buildBasicAuthorizationHeader(String username, String password) {
        String authPair = username + ":" + password;
        return "Basic " + new String(Base64.encodeBase64(authPair.getBytes()));
    }

    /**
     * Gets a username password pair from an HTTP authorization header value with
     * basic authentication method.
     *
     * @param headerValue The HTTP authorization header value
     * @return The username password pair as Array of length 2
     */
    public static String[] getBasicAuthorizationPair(String headerValue) {
        StringTokenizer st = new StringTokenizer(headerValue);
        String basic = st.nextToken();

        if (basic == null || ! basic.equalsIgnoreCase("basic")) {
            return new String[] { null, null };
        }

        String credentials = st.nextToken();
        String userPass = new String(Base64.decodeBase64(credentials));
        int pivot = userPass.indexOf(":");
        if (pivot == -1 ) {
            return new String[] { null, null };
        }

        String user = userPass.substring(0, pivot);
        String pass = userPass.substring(pivot+1);

        return new String[] { user, pass };
    }

    /**
     * Gets a content type pair, extracting the content media type
     * and content charset from a Content-Type header.
     *
     * @param contentTypeHeader The Content-Type header
     * @return The content type pair
     */
    public static String[] getContentTypePair(String contentTypeHeader) {
        if (contentTypeHeader == null || contentTypeHeader.isEmpty()) {
            return new String[] { null, null };
        }
        if (! contentTypeHeader.contains(";")) {
            return new String[] { contentTypeHeader.trim() , null };
        }
        int pivot = contentTypeHeader.indexOf(";");
        String mediaType = contentTypeHeader.substring(0, pivot).trim();
        String charsetCandidate = contentTypeHeader.substring(pivot+1).trim();

        String charset = null;

        if (FIND_CHARSET.matcher(charsetCandidate).matches()) {
            charset = FIND_CHARSET.matcher(charsetCandidate).replaceAll("$1");
        } else if (FIND_NO_EQUAL_CONTAINED.matcher(charsetCandidate).matches()) {
            charset = charsetCandidate;
        }
        return new String[] { mediaType, charset };
    }

    /**
     * Extracts a the specified header attribute from the given header value.
     *
     * @param headerValue The header value
     * @param attributeName The attribute name
     * @return The attribute value
     */
    public static String extractHeaderValueAttribute(String headerValue, String attributeName) {
        String cleanedHeaderValue = headerValue.replaceAll(" ", "");
        Matcher matcher = FIND_ATTRIBUTES.matcher(cleanedHeaderValue);
        while(matcher.find()) {
            String foundAttributeName = matcher.group(1);
            if (attributeName.equals(foundAttributeName)) {
                return matcher.group(2);
            }
        }
        return null;
    }

    /**
     * Eliminates the attributes of the given header value.
     *
     * @param headerValue The header value
     * @return The header value without attributes
     */
    public static String eliminateHeaderValueAttributes(String headerValue) {
        return FIND_SEMI_COLON_TO_END.matcher(headerValue).replaceAll("");
    }

    /**
     * Eliminates the attributes of the given header values.
     *
     * @param headerValues The header values
     * @return The header values without attributes
     */
    public static String[] eliminateHeaderValuesAttributes(String[] headerValues) {
        List<String> cleanHeaderValues = new ArrayList<String>(headerValues.length);
        for (String headerValue : headerValues) {
            String cleanHeaderValue = FIND_SEMI_COLON_TO_END.matcher(headerValue).replaceAll("");
            cleanHeaderValues.add(cleanHeaderValue);
        }
        return cleanHeaderValues.toArray(new String[cleanHeaderValues.size()]);
    }

    /**
     * Concatenates the given header values as a single String separating the
     * values by commas.
     *
     * @param headerValues The header values
     * @return The concatenated header values
     */
    public static String concatenateHeaderValues(String[] headerValues) {
        if (headerValues.length == 0) {
            return null;
        } else if (headerValues.length == 1) {
            return headerValues[0];
        }
        StringBuilder sb = new StringBuilder();
        for (String headerValue : headerValues) {
            if (sb.length() != 0) {
                sb.append(",");
            }
            sb.append(headerValue);
        }
        return sb.toString();
    }

    /**
     * Splits the comma separated header values into single values.
     *
     * @param headerValues The comma-separated header values
     * @return The single header values as an array
     */
    public static String[] splitHeaderValues(String headerValues) {
        return headerValues.split("[ ]*,[ ]*");
    }

    /**
     * Converts an input map holding its values as a list into an output map holding
     * its values as an array.
     *
     * @param inputMap The input map
     * @return The output map
     */
    public static Map<String, String[]> normalizeHeaderMap(Map<String, List<String>> inputMap) {
        Map<String, String[]> normalizedMap = new CaseInsensitiveMap<String[]>();
        for (String headerName : inputMap.keySet()) {
            List<String> headerValueList = inputMap.get(headerName);
            String[] headerValues = headerValueList.toArray(new String[headerValueList.size()]);
            normalizedMap.put(headerName, headerValues);
        }
        return normalizedMap;
    }
}
