/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply;

import net.craftforge.essential.supply.exceptions.SupplyException;
import net.craftforge.essential.supply.exceptions.UnsupportedMediaTypeException;
import net.craftforge.reflection.managers.ClassManager;

import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

/**
 * Utilities for supplier actions.
 *
 * @author Christian Bick
 * @since 25.10.11
 */
public class SupplyReflUtils {

    /**
     * Designates the producer from the list of producer classes
     * which @Produces annotation value best matches the given media types.
     *
     * @param producerClasses The list of producer classes
     * @param mediaTypes The media types
     * @return The method matching best the annotation requirements
     * @throws UnsupportedMediaTypeException if none of the given media types is supported
     */
    public static Class<?> designateProducer(List<Class<?>> producerClasses, String[] mediaTypes) throws UnsupportedMediaTypeException {
        return designateSupplier(producerClasses, Produces.class, mediaTypes);
    }

    /**
     * Designates the consumer from the list of consumer classes
     * which @Consumes annotation value best matches the given media types.
     *
     * @param consumerClasses The list of consumer classes
     * @param mediaTypes The media types
     * @return The method matching best the annotation requirements
     * @throws net.craftforge.essential.supply.exceptions.SupplyException Failed to find a consumer method supplying one of the accepted media types
     * @throws UnsupportedMediaTypeException if none of the given media types is supported
     */
    public static Class<?> designateConsumer(List<Class<?>> consumerClasses, String[] mediaTypes) throws UnsupportedMediaTypeException {
        return designateSupplier(consumerClasses, Consumes.class, mediaTypes);
    }

    /**
     * Designates the supplier(consumer or producer) from the list of supplier classes
     * which @Consumes / @Produces annotation value best matches the given media types.
     *
     * @param supplierClasses The list of supplier classes
     * @param annotationType The annotation type.
     * @param mediaTypes The media types
     * @return The method matching best the annotation requirements
     * @throws UnsupportedMediaTypeException if none of the given media types is supported
     */
    public static Class<?> designateSupplier(List<Class<?>> supplierClasses, Class<? extends Annotation> annotationType, String[] mediaTypes) throws UnsupportedMediaTypeException {
        int maxQuality = 0;
        Class<?> bestSupplierClass = null;
        for (Class<?> supplierClass : supplierClasses) {
            if (! supplierClass.isAnnotationPresent(annotationType)) {
                continue;
            }
            Annotation annotation = supplierClass.getAnnotation(annotationType);
            String[] value;
            if (annotationType.equals(Produces.class)) {
                Produces produces = (Produces)annotation;
                value = produces.value();
            } else if (annotationType.equals(Consumes.class)) {
                Consumes consumes = (Consumes)annotation;
                value = consumes.value();
            } else {
                throw new SupplyException("Unsupported annotation type: " + annotationType);
            }
            int matchingQuality = getMatchingQuality(value, mediaTypes);
            if (matchingQuality > maxQuality) {
                bestSupplierClass = supplierClass;
                maxQuality = matchingQuality;
            }
        }
        if (bestSupplierClass == null) {
            throw new UnsupportedMediaTypeException("No known supplier does support any of the given " +
                    "media types");
        }
        return bestSupplierClass;
    }

    /**
     * Designates the media type of the supplier(consumer or producer) from the list of supplier classes
     * which @Consumes / @Produces annotation value best matches the given media types.
     *
     * @param supplierClasses The list of suppliers
     * @param annotationType The annotation type.
     * @param mediaTypes The media types
     * @return The media type matching best the annotation requirements
     * @throws UnsupportedMediaTypeException if none of the given media types is supported
     */
    public static String designateBestMatchingMediaType(List<Class<?>> supplierClasses, Class<? extends Annotation> annotationType, String[] mediaTypes) throws UnsupportedMediaTypeException {
        int maxQuality = 0;
        String bestMediaType = null;
        for (Class<?> supplierClass : supplierClasses) {
            if (! supplierClass.isAnnotationPresent(annotationType)) {
                continue;
            }
            Annotation annotation = supplierClass.getAnnotation(annotationType);
            String[] value;
            if (annotationType.equals(Produces.class)) {
                Produces produces = (Produces)annotation;
                value = produces.value();
            } else if (annotationType.equals(Consumes.class)) {
                Consumes consumes = (Consumes)annotation;
                value = consumes.value();
            } else {
                throw new SupplyException("Unsupported annotation type: " + annotationType);
            }
            for (String singleValue : value) {
                int matchingQuality = getMatchingQuality(new String[] {singleValue}, mediaTypes);
                if (matchingQuality > maxQuality) {
                    bestMediaType = singleValue;
                    maxQuality = matchingQuality;
                }
            }
        }
        if (bestMediaType == null) {
            throw new UnsupportedMediaTypeException("No known supplier does support any of the given " +
                    "media types");
        }
        return bestMediaType;
    }

    /**
     * Gets the matching quality of an acceptedTypes and an available media type enumeration.
     * The quality depends on how good the best of the available media types matches against
     * the acceptedTypes media type. The higher the number, the better is the quality.<br>
     * <br>
     * 0: None of the available media types matches.<br>
     * 1: The best match was found at &#42;/&#42; vs. static/static<br>
     * 2: The best match was found at &#42;/&#42; vs. static/&#42;<br>
     * 3: The best match was found at &#42;/&#42; vs. &#42;/&#42;<br>
     * 4: The best match was fount at static/&#42; vs static/static<br>
     * 5: The best match was found at static/&#42; vs. static/&#42;<br>
     * 6: The best match was found at static/static vs. static/static (exact match)<br>
     *
     * @param availableTypes The comma separated enumeration of available media types
     * @param acceptedTypes The comma separated enumeration of accepted media types
     * @return The quality as a figure between 0 (lowest) and 6 (highest)
     */
    public static int getMatchingQuality(String availableTypes[], String acceptedTypes[]) {
        int maxQuality = 0;
        for (String acceptedType : acceptedTypes) {
            acceptedType = acceptedType.trim();
            String[] acceptedParts = acceptedType.split("/");
            if (acceptedParts.length == 1) {
                acceptedParts = new String[] { acceptedParts[0], "*" };
            }
            for (String availableType : availableTypes) {
                availableType = availableType.trim();
                String[] availableParts = availableType.split("/");
                if (availableParts.length == 1) {
                    availableParts = new String[] { availableParts[0], "*" };
                }
                int currentQuality = 0;
                if (availableParts[0].equals("*") && availableParts[1].equals("*")) {
                    if (acceptedParts[0].equals("*") && acceptedParts[1].equals("*")) {
                        currentQuality = 3;
                    } else if (acceptedParts[1].equals("*")) {
                        currentQuality = 2;
                    } else {
                        currentQuality = 1;
                    }
                } else if (availableParts[0].equals(acceptedParts[0]) && availableParts[1].equals("*")) {
                    if (acceptedParts[1].equals("*")) {
                        currentQuality = 5;
                    } else {
                        currentQuality = 4;
                    }
                } else if (availableParts[0].equals(acceptedParts[0]) && availableParts[1].equals(acceptedParts[1])) {
                    currentQuality = 6;
                }
                maxQuality = Math.max(currentQuality, maxQuality);
            }
        }
        return maxQuality;
    }

    /**
     * Gets the supported producer media types from a list of producer classes.
     *
     * @param classes The producer classes
     * @return The supported media types
     */
    public static List<String> getSupportedProducerMediaTypes(List<Class<?>> classes) {
        List<String> mediaTypes = new LinkedList<String>();
        for (Class<?> clazz : classes) {
            Produces ann = (Produces)ClassManager.getInstance(clazz).getTypeLevelAnnotation(Produces.class);
            mediaTypes.addAll(Arrays.asList(ann.value()));
        }
        return mediaTypes;
    }

    /**
     * Gets the supported consumer media types from a list of consumer classes.
     *
     * @param classes The consumer classes
     * @return The supported media types
     */
    public static List<String> getSupportedConsumerMediaTypes(List<Class<?>> classes) {
        List<String> mediaTypes = new LinkedList<String>();
        for (Class<?> clazz : classes) {
            Consumes ann = (Consumes)ClassManager.getInstance(clazz).getTypeLevelAnnotation(Consumes.class);
            mediaTypes.addAll(Arrays.asList(ann.value()));
        }
        return mediaTypes;
    }
}
