/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply.designators;

import net.craftforge.commons.collections.CollectionsUtils;
import net.craftforge.essential.supply.Consumer;
import net.craftforge.essential.supply.Consumes;
import net.craftforge.essential.supply.SupplyReflUtils;
import net.craftforge.essential.supply.exceptions.SupplyException;
import net.craftforge.essential.supply.exceptions.UnsupportedMediaTypeException;
import net.craftforge.reflection.utils.PackageUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

/**
 * A consumer designator determines the (best) consumer to
 * choose from collection of consumer classes compared
 * against a collection of desired media types.
 *
 * @author Christian Bick
 * @since 18.11.11
 */
public class ConsumerDesignator {

    /**
     * The default package where to look up consumer classes
     */
    public static final String DEFAULT_PACKAGE = "net.craftforge.essential.supply.suppliers";

    private static final Logger LOGGER = LoggerFactory.getLogger(ConsumerDesignator.class);

    private static ConcurrentHashMap<String, ConsumerDesignator> packageInstances = new ConcurrentHashMap<String, ConsumerDesignator>();

    private static ConcurrentHashMap<String, ConsumerDesignator> classListInstances = new ConcurrentHashMap<String, ConsumerDesignator>();

    /**
     * Initializes a consumer designator using the default package
     * as source for the consumer classes to use.
     *
     * @return The consumer designator
     */
    public static ConsumerDesignator getInstance() {
        return getInstance(DEFAULT_PACKAGE);
    }

    /**
     * Initializes a consumer designator using the consumer classes
     * found in the given package specified by package name.
     *
     * @param packageName The package
     * @return The consumer designator
     */
    public static ConsumerDesignator getInstance(String packageName) {
        if (! packageInstances.containsKey(packageName)) {
            LOGGER.info("[Consumer designator initialization from package] {} ", packageName);
            List<Class<?>> consumerClasses = PackageUtils.findClasses(packageName, Consumes.class);
            packageInstances.putIfAbsent(packageName, new ConsumerDesignator(consumerClasses));
        }
        return packageInstances.get(packageName);
    }

    /**
     * Initializes a consumer designator using the consumer classes
     * specified.
     *
     * @param consumerClasses The consumer classes
     * @return The consumer designator
     */
    public static ConsumerDesignator getInstance(List<Class<?>> consumerClasses) {
        String key = CollectionsUtils.buildKeyFromClassList(consumerClasses);
        if (! classListInstances.containsKey(key)) {
            LOGGER.info("[Consumer designator initialization from class list] {} ", key);
            classListInstances.putIfAbsent(key, new ConsumerDesignator(consumerClasses));
        }
        return classListInstances.get(key);
    }

    private List<Class<?>> consumerClasses;

    private ConsumerDesignator(List<Class<?>> consumerClasses) {
        this.consumerClasses = consumerClasses;
    }

    /**
     * Designates the consumer best matching the given media types.
     *
     * @param mediaTypes The list of comma separated media types
     * @return The best matching consumer
     * @throws UnsupportedMediaTypeException if no consumer supports any of the given media types
     */
    public Consumer designate(String[] mediaTypes) throws UnsupportedMediaTypeException {
        Class<?> consumerClass = SupplyReflUtils.designateConsumer(consumerClasses, mediaTypes);
        Consumer consumer;
        try {
            consumer = (Consumer)consumerClass.newInstance();
        } catch (InstantiationException e) {
            throw new SupplyException("Unable to instantiate consumer " + consumerClass.getName(), e);
        } catch (IllegalAccessException e) {
            throw new SupplyException("Unable to instantiate consumer " + consumerClass.getName(), e);
        }
        return consumer;
    }
}
