/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply.designators;

import net.craftforge.commons.collections.CollectionsUtils;
import net.craftforge.essential.supply.Producer;
import net.craftforge.essential.supply.Produces;
import net.craftforge.essential.supply.SupplyReflUtils;
import net.craftforge.essential.supply.exceptions.SupplyException;
import net.craftforge.essential.supply.exceptions.UnsupportedMediaTypeException;
import net.craftforge.reflection.utils.PackageUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

/**
 * A producer designator determines the (best) producer to
 * choose from collection of producer classes compared
 * against a collection of desired media types.
 *
 * @author Christian Bick
 * @since 18.11.11
 */
public class ProducerDesignator {

    /**
     * The default package where to look up producer classes
     */
    public static final String DEFAULT_PACKAGE = "net.craftforge.essential.supply.suppliers";

    private static final Logger LOGGER = LoggerFactory.getLogger(ProducerDesignator.class);

    private static ConcurrentHashMap<String, ProducerDesignator> packageInstances = new ConcurrentHashMap<String, ProducerDesignator>();

    private static ConcurrentHashMap<String, ProducerDesignator> classListInstances = new ConcurrentHashMap<String, ProducerDesignator>();

    /**
     * Initializes a producer designator using the default package
     * as source for the producer classes to use.
     *
     * @return The producer designator
     */
    public static ProducerDesignator getInstance() {
        return getInstance(DEFAULT_PACKAGE);
    }

    /**
     * Initializes a producer designator using the producer classes
     * found in the given package specified by package name.
     *
     * @param packageName The package
     * @return The producer designator
     */
    public static ProducerDesignator getInstance(String packageName) {
        if (! packageInstances.containsKey(packageName)) {
            LOGGER.info("[Producer designator initialization from package] {} ", packageName);
            List<Class<?>> producerClasses = PackageUtils.findClasses(packageName, Produces.class);
            packageInstances.putIfAbsent(packageName, new ProducerDesignator(producerClasses));
        }
        return packageInstances.get(packageName);
    }

    /**
     * Initializes a producer designator using the producer classes
     * specified.
     *
     * @param producerClasses The producer classes
     * @return The producer designator
     */
    public static ProducerDesignator getInstance(List<Class<?>> producerClasses) {
        String key = CollectionsUtils.buildKeyFromClassList(producerClasses);
        if (! classListInstances.containsKey(key)) {
            LOGGER.info("[Producer designator initialization from class list] {} ", key);
            classListInstances.putIfAbsent(key, new ProducerDesignator(producerClasses));
        }
        return classListInstances.get(key);
    }

    private final List<Class<?>> producerClasses;

    private ProducerDesignator(List<Class<?>> producerClasses) {
        this.producerClasses = producerClasses;
    }

    /**
     * Designates the producer best matching the given media types.
     *
     * @param mediaTypes The list of comma separated media types
     * @return The best matching producer
     * @throws UnsupportedMediaTypeException if no producer supports any of the given media types
     */
    public Producer designate(String[] mediaTypes) throws UnsupportedMediaTypeException {
        Class<?> producerClass = SupplyReflUtils.designateProducer(producerClasses, mediaTypes);
        Producer producer;
        try {
            producer = (Producer)producerClass.newInstance();
        } catch (InstantiationException e) {
            throw new SupplyException("Unable to instantiate producer " + producerClass.getName(), e);
        } catch (IllegalAccessException e) {
            throw new SupplyException("Unable to instantiate producer " + producerClass.getName(), e);
        }
        return producer;
    }

/**
     * Designates the producer best matching the given media types.
     *
     * @param mediaTypes The media types
     * @return The best matching producer
     * @throws UnsupportedMediaTypeException if no producer supports any of the given media types
     */
    public String designateMediaType(String[] mediaTypes) throws UnsupportedMediaTypeException {
        return SupplyReflUtils.designateBestMatchingMediaType(producerClasses, Produces.class, mediaTypes);
    }

}
