/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply.suppliers.basic;

import net.craftforge.commons.io.StreamUtils;
import net.craftforge.essential.supply.Consumer;
import net.craftforge.essential.supply.Consumes;
import net.craftforge.essential.supply.exceptions.BadInputException;
import net.craftforge.essential.supply.exceptions.SupplyException;
import net.craftforge.reflection.utils.ReflUtils;

import java.io.Reader;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * <p>A basic consumer is used to interpret input as plain text and to convert it to
 * real or virtual primitives.</p>
 * <p>Virtual primitives are all object representations of real primitives
 * (e.g. Integer) as well as types like String or Date.</p>
 *
 * @author Christian Bick
 * @since 18.11.11
 */
@Consumes({"text/plain"})
public class BasicConsumer implements Consumer {

    /**
     * Consumes the given reader interpreting the read content as plain text, and deserializing it to an
     * object of the specified class.
     *
     * @param clazz The class of the desired result object
     * @param reader The reader
     * @return The result object
     * @throws BadInputException if the input cannot be unmarshalled
     */
    public Object consume(Class<?> clazz, Reader reader) throws BadInputException, SupplyException {
        String input = StreamUtils.readStringFromReader(reader);

        if (clazz.isPrimitive()) {
            clazz = ReflUtils.primitiveToObject(clazz);
        }
        try {
            Constructor constructor = clazz.getConstructor(new Class[] {String.class});
            return constructor.newInstance(input);
        } catch (NoSuchMethodException e) {
            try {
                Method method = clazz.getMethod("valueOf", new Class[] {String.class});
                return method.invoke(null, input);
            } catch (NoSuchMethodException e1) {
                throw new SupplyException("Unable to find a method to initialize object from input", e);
            } catch (InvocationTargetException e1) {
                throw new SupplyException(e);
            } catch (IllegalAccessException e1) {
                throw new SupplyException(e);
            }
        } catch (InvocationTargetException e) {
            Throwable cause = e.getCause();
            if (cause instanceof NumberFormatException) {
                throw new BadInputException("Unable to interpret the value " + input +
                        " as " + clazz.getSimpleName(), cause);
            }
            throw new SupplyException(e);
        } catch (InstantiationException e) {
            throw new SupplyException(e);
        } catch (IllegalAccessException e) {
            throw new SupplyException(e);
        }
    }
}
