/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply.suppliers.jackson;

import net.craftforge.essential.supply.Consumer;
import net.craftforge.essential.supply.Consumes;
import net.craftforge.essential.supply.exceptions.BadInputException;
import net.craftforge.essential.supply.exceptions.SupplyException;
import org.codehaus.jackson.JsonParseException;
import org.codehaus.jackson.map.DeserializationConfig;
import org.codehaus.jackson.map.JsonMappingException;
import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.xc.JaxbAnnotationIntrospector;

import java.io.IOException;
import java.io.Reader;

/**
 * A jackson consumer deserializes JSON to objects. This process can
 * be fine tuned via configuration.
 *
 * @author Christian Bick
 * @since 18.11.11
 */
@Consumes({"text/json", "application/json"})
public class JacksonConsumer implements Consumer {

    private JacksonConfiguration config;

    /**
     * Initializes a Jackson consumer using the configuration from
     * the configuration's default file if found in the class path.
     *
     * @see JacksonConfiguration
     */
    public JacksonConsumer() {
        this.config = new JacksonConfiguration();
    }

    /**
     * Initializes a Jackson consumer using the given configuration.
     *
     * @param config The configuration
     */
    public JacksonConsumer(JacksonConfiguration config) {
        this.config = config;
    }

    /**
     * Deserializes the reader content given in JSON format to an object of the specified type.
     *
     * @param type The type
     * @param reader The reader
     * @return The object
     * @throws BadInputException if the input cannot be deserialized
     */
    public Object consume(Class<?> type, Reader reader) throws BadInputException, SupplyException {
        try {
            ObjectMapper mapper = new ObjectMapper();
            mapper.getDeserializationConfig().withAnnotationIntrospector(new JaxbAnnotationIntrospector());
            for (String key : config.getJsonDeserializationProperties().keySet()) {
                String value = config.getJsonDeserializationProperties().get(key);
                mapper.getDeserializationConfig().set(DeserializationConfig.Feature.valueOf(key), Boolean.valueOf(value));
            }
            Object value;
            value = mapper.readValue(reader, type);
            reader.close();
            return value;
        } catch (JsonParseException e) {
            throw new BadInputException("Failed to parse the given JSON", e);
        } catch (JsonMappingException e) {
            throw new BadInputException("Failed to map the given JSON to class " + type, e);
        } catch (IOException e) {
            throw new SupplyException("Failed to deserialize the given JSON", e);
        }
    }
}
