/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply.suppliers.jackson;

import net.craftforge.essential.supply.Producer;
import net.craftforge.essential.supply.Produces;
import net.craftforge.essential.supply.exceptions.SupplyException;
import org.codehaus.jackson.JsonGenerator;
import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.map.SerializationConfig;
import org.codehaus.jackson.xc.JaxbAnnotationIntrospector;

import java.io.IOException;
import java.io.Writer;
import java.util.Map;

/**
 * A jackson producer serializes objects to a JSON. This process can
 * be fine tuned via configuration.
 *
 * @author Christian Bick
 * @since 18.11.11
 */
@Produces({"text/json", "application/json"})
public class JacksonProducer implements Producer {

    private JacksonConfiguration config;

    /**
     * Initializes a Jackson producer using the configuration from
     * the configuration's default file if found in the class path.
     *
     * @see JacksonConfiguration
     */
    public JacksonProducer() {
        this.config = new JacksonConfiguration();
    }

    /**
     * Initializes a Jackson producer using the given configuration.
     *
     * @param config The configuration
     */
    public JacksonProducer(JacksonConfiguration config) {
        this.config = config;
    }

    /**
     * Serializes the given object to JSON using the specified writer.
     *
     * @param object The object
     * @param writer The writer
     */
    public void produce(Object object, Writer writer) throws SupplyException {
        try {
            ObjectMapper mapper = new ObjectMapper();
            mapper.getDeserializationConfig().withAnnotationIntrospector(new JaxbAnnotationIntrospector());
            for (Map.Entry<String, String> entry : config.getJsonSerializationProperties().entrySet()) {
                String key = entry.getKey();
                String value = entry.getValue();
                mapper.getSerializationConfig().set(SerializationConfig.Feature.valueOf(key), Boolean.valueOf(value));
            }
            JsonGenerator generator = null;
            generator = mapper.getJsonFactory().createJsonGenerator(writer);

            generator.writeObject(object);
            generator.close();
        } catch (IOException e) {
            throw new SupplyException("Failed to serialize the given object to JSON", e);
        }
    }
}
