/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply.suppliers.jaxb;

import net.craftforge.essential.supply.Consumer;
import net.craftforge.essential.supply.Consumes;
import net.craftforge.essential.supply.exceptions.BadInputException;
import net.craftforge.essential.supply.exceptions.SupplyException;

import javax.xml.bind.*;
import java.io.IOException;
import java.io.Reader;

/**
 * A JAXB consumer deserializes XML to objects. This process can
 * be fine tuned via configuration.
 *
 * @author Christian Bick
 * @since 26.02.2011
 */
@Consumes({"text/xml", "application/xml"})
public class JaxbConsumer implements Consumer {

    private JaxbConfiguration config;

    /**
     * Initializes a JAXB consumer using the configuration from
     * the configuration's default file if found in the class path.
     *
     * @see JaxbConfiguration
     */
    public JaxbConsumer() {
        this.config = new JaxbConfiguration();
    }

    /**
     * Initializes a JAXB consumer using the given configuration.
     *
     * @param config The configuration
     */
    public JaxbConsumer(JaxbConfiguration config) {
        this.config = config;
    }

    /**
     * Deserializes the reader content given in XML format to an object of the specified clazz.
     *
     * @param clazz The clazz
     * @param reader The reader
     * @return The object
     * @throws BadInputException if the input cannot be unmarshalled
     */
    public Object consume(Class<?> clazz, Reader reader) throws BadInputException, SupplyException {
        try {
            JAXBContext jc = JAXBContext.newInstance(clazz);
            Unmarshaller u = jc.createUnmarshaller();
            for (String key : config.getXmlDeserializationProperties().keySet()) {
                String value = config.getXmlDeserializationProperties().get(key);
                u.setProperty(key, value);
            }
            Object value;
            value = u.unmarshal(reader);
            reader.close();
            return value;
        } catch (UnmarshalException e) {
            throw new BadInputException("Failed to unmarshall the given XML", e);
        } catch (PropertyException e) {
            throw new SupplyException("Failed to set JAXB properties", e);
        } catch (JAXBException e) {
            throw new SupplyException("Failed to set up JAXB properly", e);
        } catch (IOException e) {
            throw new SupplyException("Failed to perform IO operation", e);
        }
    }

}
