/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.supply.suppliers.jaxb;

import net.craftforge.essential.supply.Producer;
import net.craftforge.essential.supply.Produces;
import net.craftforge.essential.supply.exceptions.SupplyException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.PropertyException;
import java.io.IOException;
import java.io.Writer;
import java.util.Map;

/**
 * A JAXB producer serializes objects to a XML. This process can
 * be fine tuned via configuration.
 * 
 * @author Christian Bick
 * @since 02.12.2010
 */
@SuppressWarnings("unused")
@Produces({"text/xml", "application/xml"})
public class JaxbProducer implements Producer {

    private JaxbConfiguration config;

    /**
     * Initializes a JAXB producer using the configuration from
     * the configuration's default file if found in the class path.
     *
     * @see JaxbConfiguration
     */
    public JaxbProducer() {
        this.config = new JaxbConfiguration();
    }

    /**
     * Initializes a JAXB producer using the given configuration.
     *
     * @param config The configuration
     */
    public JaxbProducer(JaxbConfiguration config) {
        this.config = config;
    }

    /**
     * Serializes the given object to XML using the specified writer.
     *
     * @param object The object
     * @param writer The writer
     */
    public void produce(Object object, Writer writer) throws SupplyException {
        try{
            JAXBContext context = JAXBContext.newInstance(object.getClass());
            Marshaller marshaller = context.createMarshaller();
            for (Map.Entry<String, String> entry : config.getXmlSerializationProperties().entrySet()) {
                String key = entry.getKey();
                String value = entry.getValue();
                marshaller.setProperty(key, convertPropertyValue(key, value));
            }
            marshaller.marshal(object, writer);
            writer.close();
        } catch (PropertyException e) {
            throw new SupplyException("Failed to set JAXB properties", e);
        } catch (JAXBException e) {
            throw new SupplyException("Failed to set up JAXB properly", e);
        } catch (IOException e) {
            throw new SupplyException("Failed to perform IO operation", e);
        }
    }

    /**
     * Converts a JAXB Marshaller property value into the desired boolean for
     * boolean Marshaller properties.
     *
     * @param propertyName The property name
     * @param propertyValue The property value
     * @return The converted property value
     */
    private Object convertPropertyValue(String propertyName, String propertyValue) {
        if (propertyName.equals(Marshaller.JAXB_FORMATTED_OUTPUT) || propertyName.equals(Marshaller.JAXB_FRAGMENT)) {
            return Boolean.valueOf(propertyValue);
        }
        return propertyValue;
    }
}
